;;;; Ant minor mode - work with Ant to build any project
;;;; Copyright (C) 2001 Michel CASABIANCA (casa@sweetohm.net) and
;;;;                    Frdric Brunel (brunel@dotcom.fr)
;;;;
;;;;   This program is free software; you can redistribute it and/or
;;;;   modify it under the terms of the GNU General Public License
;;;;   as published by the Free Software Foundation; either version 2
;;;;   of the License, or (at your option) any later version.    
;;;;
;;;;   This program is distributed in the hope that it will be useful,
;;;;   but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;;   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;;   GNU General Public License for more details.
;;;;
;;;;   A copy of the GNU General Public License is included with the
;;;;   GNU Emacs distribution.
;;;;
;;;; Ant is a build tool (like make) that proves to be very useful in
;;;; many situations: you may use it to build your Java projects, but
;;;; also to build a web site or documentations. Thus, I made it a 
;;;; minor mode that may live with JDE or PSGML.
;;;;
;;;; INSTALL
;;;;
;;;; o First drop this file in the directory of your choice (which is 
;;;;   "/usr/local/emacs/lisp" in the following configuration sample).
;;;; o If you are running Emacs 20, you will need the file "xml.el"
;;;;   that you can download with "ant.el". Put it in the same directory.
;;;;   This file is included in the Emacs 21 distribution.
;;;; o Then byte compile the .el file(s) to get .elc files(s) that are
;;;;   faster to load.
;;;;   - If you run Emacs 20, type on the command line:
;;;;     emacs -batch -q -l xml.el -f batch-byte-compile ant.el xml.el
;;;;   - If you run Emacs 21, type:
;;;;     emacs -batch -q -f batch-byte-compile ant.el
;;;; o Edit your .emacs file to add the following lines:
;;;;   (add-to-list 'load-path "/usr/local/emacs/lisp")
;;;;   (load "ant")
;;;;   (ant-mode)
;;;;
;;;; Delete the last line if you don't want to enable Ant mode in all
;;;; buffers (you may then enable/disable it typing M-x ant-mode).
;;;; When this mode is enabled, you'll see (XXX Ant) in your mode line
;;;; and a new menu "Ant" in you menu bar.
;;;;
;;;; USAGE
;;;;
;;;; This mode implements the following features:
;;;;
;;;; o Build choosing the Ant target by typing C-x C-a b or selecting
;;;;   "Build" in the menu. You will be prompt for a target on the
;;;;   minibuffer. You can browse the history using arrow keys.
;;;; o Rebuild with the last used target by typing C-x C-a r or
;;;;   select "Rebuild" in the menu.
;;;; o Scan the build file to search for targets (the list of targets
;;;;   in the build file will be displayed in the minibuffer and loaded
;;;;   in the target history. The build file is search recursively in
;;;;   the file system ("build.xml" is search in the directory of the
;;;;   buffer, then in it's parent, and so on). Perform this by typing
;;;;   C-x C-a t or select "Targets" in the menu.
;;;; o Select a file a given build file. You will be prompt in the
;;;;   minibuffer for a build file. When you will build or rebuild,
;;;;   this file will be used instead of recursive search. To forget
;;;;   this file and search in the file system, type an empty string
;;;;   for the build file.
;;;;
;;;; You may customize the call to Ant with the following variables:
;;;;
;;;; o ant-command [default value: ant]: the command to launch Ant.
;;;; o ant-command-options [-emacs]: the command line options.
;;;; o ant-buildfile-name [build.xml]: the build file to search.
;;;; o ant-buildfile-path [nil]: the complete file name of the build
;;;;   file. If this file name is non nil, it is used for build and
;;;;   not searched in the file system.
;;;;
;;;; To set a given value for a variable, write in your .emacs (or
;;;; evalute) for instance:
;;;;
;;;;   (setq ant-command-options "-emacs -quiet")
;;;;
;;;; HISTORY
;;;;
;;;; o version 0.3 [2002-05-02]: New server mode implemented.
;;;; o version 0.2 [2002-02-07]: Documentation improved (installation).
;;;; o version 0.1 [2001-12-02]: XML parsing added.
;;;; o Version 0.0 [2001-12-01]: first version, quite poor.

;; needed libraries
(require 'xml)
;; Ant mode variable
(defvar ant-mode nil)
;; Add in minor-mode-alist
(if (not (assq 'ant-mode minor-mode-alist))
    (setq minor-mode-alist
	  (append minor-mode-alist '((ant-mode " Ant"))) ) )
;; Ant mode key map
(defvar ant-mode-map (make-sparse-keymap))
(define-key ant-mode-map "\C-x\C-ab" 'ant-build)
(define-key ant-mode-map "\C-x\C-ar" 'ant-rebuild)
(define-key ant-mode-map "\C-x\C-af" 'ant-buildfile)
(define-key ant-mode-map "\C-x\C-at" 'ant-targets)
(define-key ant-mode-map "\C-x\C-ah" 'ant-help)
(define-key ant-mode-map "\C-x\C-as" 'ant-server-start)
(define-key ant-mode-map "\C-x\C-ae" 'ant-server-stop)
;; Add in minor-mode-map-alist
(if (not (assq 'ant-mode minor-mode-map-alist))
    (setq minor-mode-map-alist
	  (cons (cons 'ant-mode ant-mode-map)
		minor-mode-map-alist)))
;;; Ant mode function
(defun ant-mode (&optional arg)
  "Ant mode is a minor mode to work with Ant."
  (interactive)
  (setq ant-mode
	(if (null arg) (not ant-mode)
	  (> (prefix-numeric-value arg) 0))))
;; Ant menu description
(defvar ant-menu-definition
  (list "Ant"
	["Build" ant-build t]
	["Rebuild" ant-rebuild t]
	"-"
	["Build file" ant-buildfile t]
	["Targets" ant-targets t]
	["Help" ant-help t]
	"-"
	["Start Server" ant-server-start t]
	["Stop Server" ant-server-stop t])
  "Menu for Ant")
;;; Add in menu bar
(easy-menu-define ant-menu ant-mode-map "Menu for Ant." ant-menu-definition)


;; ant command
(defvar ant-command "ant"
  "Command to launch Ant.")
;; ant command options
(defvar ant-command-options "-emacs"
  "Command lien option to launch Ant (-find is not necessary).")
;; buildfile name
(defvar ant-buildfile-name "build.xml"
  "Build file name (without directory).")
;; buildfile path
(defvar ant-buildfile-path nil
  "Build file path (complete name with directory). If nil, the build file
is searched recusively.")
;; Ant target list
(defvar ant-targets '(""))
;; Ant server process
(defvar ant-server-process nil)
;; Ant server process name
(defvar ant-server-process-name "ant-server")
;; The name of the buffer where the server is running
(defvar ant-server-buffer-name "*ant-server*")
;; Ant server task name
(defvar ant-server-task-name "server")
;; command line options to launch ant in server mode
(defvar ant-server-command-options ""
  "Command lien option to launch Ant in server mode.")
;; the server port
(defvar ant-server-port 5393
  "The Ant server port")

;;; compile using prompt target
(defun ant-build ()
  "Build prompting the target in the minibuffer."
  (interactive)
  (ant-targets)
  (let 
    ((target-name (read-from-minibuffer "Target: " (car ant-targets) nil nil 'ant-targets)))
    (save-some-buffers)
    ;; test if server is running
    (cond ((null ant-server-process)
	   ;; no server running
	   (setq compile-command (ant-compile-command))
	   (compile (concat compile-command " " target-name)))
	  (t
	   ;; a server is running
	   (send-server-command (concat "target " target-name "\n"))))))

;;; compile using the last target
(defun ant-rebuild ()
  "Build using the last target in history."
  (interactive)
  (save-some-buffers)
  ;; test if server is running
  (cond ((null ant-server-process)
	   ;; no server running
	 (setq compile-command (ant-compile-command))
	 (compile (concat "" compile-command " " (car ant-targets))))
	(t
	 ;; a server is running
	 (send-server-command (concat "target " (car ant-targets) "\n")))))

;;; send a command to the server
(defun send-server-command (command)
  "Send a command to the server."
  (open-network-stream "ant-client" 
		       "*compilation*" 
		       "localhost" 
		       ant-server-port)
  (process-send-string "ant-client" command))

;;; display help on targets
(defun ant-help ()
  "Display help on targets."
  (interactive)
  ;; build the help command
  (setq compile-command (ant-compile-command))
  (compile (concat "" compile-command " -projecthelp -quiet")))

;;; scan build file to fetch targets
(defun ant-targets ()
  "Scan the build file and return the list of targets."
  (interactive)
  ;; get the buildfile
  (let ((ant-buildfile (ant-get-buildfile)))
	   (cond ((not (null ant-buildfile))
		  ;; get the target nodes
		  (defvar target-nodes 
		    (remove nil 
			    (mapcar (lambda (node)
				      (when (eq (xml-node-name node) 'target)
					node))
				    (xml-node-children 
				     (car (xml-parse-file ant-buildfile))))))
                  ;; get the target list
		  (setq ant-targets
			(mapcar (lambda (node)
				  (cdr (assoc 'name 
					      (xml-node-attributes node))))
				target-nodes))
		  ;; print the target list
		  (message "Targets: %s" ant-targets)))))

;;; start ant server
(defun ant-server-start ()
  "Start the Ant server."
  (interactive)
  ;; test if server already running
  (cond ((null ant-server-process)
	 ;; start the process
	 (let ((ant-buildfile (ant-get-buildfile)))
	   (cond ((not (null ant-buildfile))
		  (start-process ant-server-process-name 
				 ant-server-buffer-name
				 ant-command
				 (if (not (null ant-server-command-options))
				     "-buildfile")
				 ant-buildfile
				 ant-server-task-name)
		  (setq ant-server-process "ant-server")
		  (switch-to-buffer "*ant-server*")))))
	(t
	 ;; server already running: error message
	 (message "Ant Server already running !")
	 (beep))))

;;; stop ant server (maybe quite violent to kill the process ?)
(defun ant-server-stop ()
  "Stop the Ant server."
  (interactive)
  ;; test if server already running
  (cond ((null ant-server-process)
	 (message "Ant Server not running !")
	 (beep))
	(t
	 ;; stop the process
	 (delete-process ant-server-process)
	 (setq ant-server-process nil))))

;;; set the buildfile to use (no recursive search)
(defun ant-buildfile (file)
  "Set the file to use to build. If non nil, the build file is not
searched recursively when building"
  (interactive "fBuild file: ")
  (setq ant-buildfile-path file)
  (if (equal ant-buildfile-path "")
      (setq ant-buildfile-path nil)))

;;; build the compile command
(defun ant-compile-command ()
  (concat "" ant-command " " ant-command-options " "
	  (cond ((equal ant-buildfile-path nil)
		 (concat "-find " ant-buildfile-name))
		(t
		 (concat "-buildfile " ant-buildfile-path)))))

;;; get the buildfile and display an error message if no buildfile was found,
;;; returns nil if this case
(defun ant-get-buildfile ()
  ;; test if buildfile was set
  (cond ((not (null ant-buildfile-path))
	 ant-buildfile-path)
	;; test if buffer owns a file
	((null buffer-file-name)
	 (message "Buffer doesn't own a file !")
	 (beep))
	(t
	 (let ((ant-file (ant-find-buildfile 
			 (file-name-directory buffer-file-name))))
	   (cond ((null ant-file)
		  (message "Build file not found !")
		  (beep))
		 (t ant-file))))))

;;; find the buildfile in the current or parent directories
(defun ant-find-buildfile (dir)
  "Finds the build file in DIR or in parent directories recursively.
Return nil if not found"
  (let ((file (find ant-buildfile-name
		    (directory-files dir) :test 'string=)))
    (if file
	(expand-file-name file dir)
      (if (not (ant-root-dir-p dir))
	  (ant-find-buildfile (concat dir "../"))))))

;;; tells if a given directory is the root
(defun ant-root-dir-p (dir)
  "Tells if DIR is root dir of the file system"
  (let ((parent (concat dir "../")))
    (cond 
     ((and
       (fboundp 'ange-ftp-ftp-name)
       (ange-ftp-ftp-name dir))
      (ange-ftp-get-file-entry parent))
     ((eq system-type 'windows-nt)
      (not (file-exists-p parent)))
     ((eq system-type 'cygwin32)
      (or (string= (file-truename dir) "/") 
	  (not (file-exists-p (file-truename dir)))))
     (t
      (or (or (not (file-readable-p dir))
	      (not (file-readable-p parent)))
	  (and 
	   (string= (file-truename dir) "/")
	   (string= (file-truename parent) "/")))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; UGLY HORRIBLE COPY/PASTE FROM PROGMODES/COMPILE.EL !!!!
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun ant-compile-internal (command error-message
				 &optional name-of-mode parser
				 error-regexp-alist name-function
				 enter-regexp-alist leave-regexp-alist
				 file-regexp-alist nomessage-regexp-alist)
  (let (outbuf)
    (save-excursion
      (or name-of-mode
	  (setq name-of-mode "Compilation"))
      (setq outbuf
	    (get-buffer-create
	     (funcall (or name-function compilation-buffer-name-function
			  (function (lambda (mode)
				      (concat "*" (downcase mode) "*"))))
		      name-of-mode)))
      (set-buffer outbuf)
      (let ((comp-proc (get-buffer-process (current-buffer))))
	(if comp-proc
	    (if (or (not (eq (process-status comp-proc) 'run))
		    (yes-or-no-p
		     (format "A %s process is running; kill it? "
			     name-of-mode)))
		(condition-case ()
		    (progn
		      (interrupt-process comp-proc)
		      (sit-for 1)
		      (delete-process comp-proc))
		  (error nil))
	      (error "Cannot have two processes in `%s' at once"
		     (buffer-name))
	      )))
      ;; In case the compilation buffer is current, make sure we get the global
      ;; values of compilation-error-regexp-alist, etc.
      (kill-all-local-variables))
    (or error-regexp-alist
	(setq error-regexp-alist compilation-error-regexp-alist))
    (or enter-regexp-alist
	(setq enter-regexp-alist compilation-enter-directory-regexp-alist))
    (or leave-regexp-alist
	(setq leave-regexp-alist compilation-leave-directory-regexp-alist))
    (or file-regexp-alist
	(setq file-regexp-alist compilation-file-regexp-alist))
    (or nomessage-regexp-alist
	(setq nomessage-regexp-alist compilation-nomessage-regexp-alist))
    (or parser (setq parser compilation-parse-errors-function))
    (let ((thisdir default-directory)
	  outwin)
      (save-excursion
	;; Clear out the compilation buffer and make it writable.
	;; Change its default-directory to the directory where the compilation
	;; will happen, and insert a `cd' command to indicate this.
	(set-buffer outbuf)
	(setq buffer-read-only nil)
	(buffer-disable-undo (current-buffer))
	(erase-buffer)
	(buffer-enable-undo (current-buffer))
	(setq default-directory thisdir)
	(insert "cd " thisdir "\n" command "\n")
	(set-buffer-modified-p nil))
      ;; If we're already in the compilation buffer, go to the end
      ;; of the buffer, so point will track the compilation output.
      (if (eq outbuf (current-buffer))
	  (goto-char (point-max)))
      ;; Pop up the compilation buffer.
      (setq outwin (display-buffer outbuf))
      (save-excursion
	(set-buffer outbuf)
	(compilation-mode name-of-mode)
	;; (setq buffer-read-only t)  ;;; Non-ergonomic.
	(set (make-local-variable 'compilation-parse-errors-function) parser)
	(set (make-local-variable 'compilation-error-message) error-message)
	(set (make-local-variable 'compilation-error-regexp-alist)
	     error-regexp-alist)
	(set (make-local-variable 'compilation-enter-directory-regexp-alist)
	     enter-regexp-alist)
	(set (make-local-variable 'compilation-leave-directory-regexp-alist)
	     leave-regexp-alist)
	(set (make-local-variable 'compilation-file-regexp-alist)
	     file-regexp-alist)
	(set (make-local-variable 'compilation-nomessage-regexp-alist)
	     nomessage-regexp-alist)
	(set (make-local-variable 'compilation-arguments)
	     (list command error-message
		   name-of-mode parser
		   error-regexp-alist name-function
		   enter-regexp-alist leave-regexp-alist
		   file-regexp-alist nomessage-regexp-alist))
        (make-local-variable 'lazy-lock-defer-on-scrolling)
        ;; This proves a good idea if the buffer's going to scroll
        ;; with lazy-lock on.
        (setq lazy-lock-defer-on-scrolling t)
	(setq default-directory thisdir
	      compilation-directory-stack (list default-directory))
	(set-window-start outwin (point-min))
	(or (eq outwin (selected-window))
	    (set-window-point outwin (point-min)))
	(compilation-set-window-height outwin)
	(if compilation-process-setup-function
	    (funcall compilation-process-setup-function))
	;; Start the compilation.
	(if (fboundp 'start-process)
	    (let* ((process-environment (cons "EMACS=t" process-environment))
		   (proc (start-process-shell-command (downcase mode-name)
						      outbuf
						      command)))
	      (set-process-sentinel proc 'compilation-sentinel)
	      (set-process-filter proc 'compilation-filter)
	      (set-marker (process-mark proc) (point) outbuf)
	      (setq compilation-in-progress
		    (cons proc compilation-in-progress)))
	  ;; No asynchronous processes available.
	  (message "Executing `%s'..." command)
	  ;; Fake modeline display as if `start-process' were run.
	  (setq mode-line-process ":run")
	  (force-mode-line-update)
	  (sit-for 0)			; Force redisplay
	  (let ((status (call-process shell-file-name nil outbuf nil "-c"
				      command)))
	    (cond ((numberp status)
		   (compilation-handle-exit 'exit status
					    (if (zerop status)
						"finished\n"
					      (format "\
exited abnormally with code %d\n"
						      status))))
		  ((stringp status)
		   (compilation-handle-exit 'signal status
					    (concat status "\n")))
		  (t
		   (compilation-handle-exit 'bizarre status status))))
	  (message "Executing `%s'...done" command)))
      (if compilation-scroll-output
	  (save-selected-window
            (select-window outwin)
            (goto-char (point-max)))))
    ;; Make it so the next C-x ` will use this buffer.
    (setq compilation-last-buffer outbuf)))
